//-------------------------------------------------------------------------------
//
//      spitter.c - Command line tool to process a file of 
//      Asterisk call "records". See Explanation ().
// 
//    Copyright (C) 2006  Mark D. Collier/Mark O'Brien
//
//    This program is free software; you can redistribute it and/or modify
//    it under the terms of the GNU General Public License as published by
//    the Free Software Foundation; either version 2 of the License, or
//    (at your option) any later version.
//
//    This program is distributed in the hope that it will be useful,
//    but WITHOUT ANY WARRANTY; without even the implied warranty of
//    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
//    GNU General Public License for more details.
//
//    You should have received a copy of the GNU General Public License
//    along with this program; if not, write to the Free Software
//    Foundation, Inc., 59 Temple Place - Suite 330, Boston, MA 02111-1307, USA.
//
//   Author: Mark D. Collier/Mark O'Brien - 08/07/2006  v1.0
//         www.securelogix.com - mark.collier@securelogix.com
//         www.hackingexposedvoip.com
//
//-------------------------------------------------------------------------------

#include "spitter.h"

int  main ( int argc, char *argv[] )
{
    signal ( SIGTERM, catch_signals );
    signal ( SIGINT, catch_signals );
    
    if ( argc < 2 ) // less than the 1 mandatory parm?
    {
        printf ( "\nError: missing mandatory parameter\n" );
        usage ();
    };
    
    //
    //  parse the command line
    //

    verbosity = 0;
    
    while (( opt = getopt ( argc, argv, "l:tvh?" ) ) != EOF)
    {
        switch (opt)
        {
            case 'l':                             // lower case letter l -- as in limit.
                tempLimitSpitterConcurrentCalls = atoi ( optarg );
                break;
            case 't':
                bTestMode = true;
                break;
            case 'v':                             //  verbose option
                verbosity++;
                break;
            case 'h':                             //  explanation & usage
            case '?':
                Explanation ();
                usage ();                         // control does not return here
                break;
        }
    }    
   
    //  getopt permutes the order of the parms in argv[] placing non-optional parms
    //  at the end of argv. optind should be the index of the 1st mandatory non-optional
    //  parm in argv[]  and there must be exactly 1 non-optional mandatory parms:
    
    if ( optind != ( argc - 1 ) )
    {
        printf ( "\nError: Less than 1 mandatory parameter\n" );
        usage ();               // control does not return here
    }
    	
    psFileOfCallRecords = argv[optind++];
    
    //  Verify Asterisk is installed & running

    if ( !bTestMode )
    {
        if ( stat ( "/var/run/asterisk.pid", &statBuf ) )
        {
            perror ( "\n\nExpecting to find file /var/run/asterisk.pid if "
                     "Asterisk is installed and running. Either Asterisk "
                     "is not installed and running or your OS does not "
                     "maintain a record of running processes in /var/run/ " );
            printf ( "\n" );
        
            CleanupAndExit ( EXIT_FAILURE );    // control does not return here
        }
    }

    //  We will eventually create a file in the /tmp directory for every call record in the
    //  input file. Files in the /tmp directory are eventually dropped (moved) into Asterisk's 
    //  /var/spoo/asterisk/outgoing folder where their presence is detected by Asterisk
    //  and acted upon. Each filename must be unique. We'll use a random number
    //  generator and incorporate the random number as a part of a file's name to
    //  assure file names are unique.
    
    srand ( time ( NULL ) );  //  seed the random # generator with TOD in sec since Epoch.
    
    //  Verify the required tmp directory and the required outgoing directory exists and that
    //  they are both r/w to spitter

    if ( bTestMode )
    {
        if ( access ( "./tmp", R_OK | W_OK ) )
        {
            perror ( "\n\nTo operate in test mode, you must have a "
                     "directory called tmp within your current directory "
                     "and the tmp directory must be r/w to spitter." );
            printf ( "\n" );
            
            CleanupAndExit ( EXIT_FAILURE );    // control does not return here
        }
        
        if ( access ( "./outgoing", R_OK | W_OK ) )
        {
            perror ( "\n\nTo operate in test mode, you must have a "
                     "directory called outgoing within your current directory "
                     "and the outgoing directory must be r/w to spitter." );
            printf ( "\n" );
            
            CleanupAndExit ( EXIT_FAILURE );    // control does not return here
        }
    }
    else
    {
        if ( access ( "/tmp", R_OK | W_OK ) )
        {
            perror ( "\n\nspitter does not have r/w access to the system /tmp "
                     "directory. " );
            printf ( "\n" );
            
            CleanupAndExit ( EXIT_FAILURE );    // control does not return here
        }
        
        if ( access ( "/var/spool/asterisk/outgoing", R_OK | W_OK ) )
        {
            perror ( "\n\nspitter does not have r/w access to Asterisk's "
                     "/var/spool/asterisk/outgoing" );
            printf ( "\n" );
            
            CleanupAndExit ( EXIT_FAILURE );    // control does not return here
        }
    }
    
    //  Verify the file of call records specified by the user on the command
    //  line actually exists and can be opened for reading
    
    hFileOfCallRecords = fopen ( psFileOfCallRecords, "r" );
    
    if ( !hFileOfCallRecords )
    {
        //  Inform user that file of call records cannot be opened
        
        perror ( "\nCannot open input file for reading." );
        printf ( "\n" );
        CleanupAndExit ( EXIT_FAILURE );     // control does not return here
    }
    
    //  Collect Call Record statistics (i.e. # records, # lines in file).
    
    bLookingForBlankLine = false;
    numCallRecords = 0;
    numLines = 0;
        
    while ( fgets ( line, 2048, hFileOfCallRecords ) != NULL )
    {
        numLines++;
        
        if ( line[0] == '\n' )   // found a blank line?
        {
            //  Blank line encountered
            
            if ( bLookingForBlankLine )
            {
                //  Found end of current call record
                
                bLookingForBlankLine = false;  //  looking for start of next call record              
            }
            else
            {
                ; //  continue scanning for start of next call record
            }
        }
        else
        {
            //  Non-blank line encountered
            
            if ( !bLookingForBlankLine )
            {
                numCallRecords++;
                bLookingForBlankLine = true;  //  scan for end of this call record

            }
            else
            {
                ; //  continue scanning for end of present call record
            }
        }
        
    } //  end while ( input call record file line != NULL )
    
    if ( tempLimitSpitterConcurrentCalls < 0 )
    {
        printf ( "\n\nLimit of concurrent spitter calls cannot be negative\n" );
        usage ();
    }

    limSpitterConcurrentCalls = tempLimitSpitterConcurrentCalls;
            
    //  Print summary of command line inputs

    printf ( "\n%s\n", __SPITTER_VERSION );
    printf ( "%s\n",   __SPITTER_DATE );
    
    printf ( "\nFile of Call Records:           %s", psFileOfCallRecords );
    printf ( "\nNumber of Lines in File:        %u", numLines );
    printf ( "\nNumber of Call Records Found:   %u\n", numCallRecords );
    
    if ( limSpitterConcurrentCalls )
    {
        printf ( "\nLimit of concurrent SPIT calls: %u\n", limSpitterConcurrentCalls );
    }
    else
    {
        printf ( "\nLimit of concurrent SPIT calls: UNLIMITED\n" );
    }
    
    if ( bTestMode )
    {
        printf ( "\nTest Mode! This is simulated SPIT!\n\n"
                 "In test mode you must have a /tmp folder and a /outgoing "
                 "folder present in your current directory. There is no "
                 "check to confirm Asterisk is running. No call files "
                 "are sent to Asterisk's actual outgoing folder. "
                 "Instead, call files are temporarily created in your "
                 "personal /tmp folder and moved from there to your "
                 "personal /outgoing folder\n" );
    }
    
    if ( verbosity )
    {
        printf ( "\nVerbosity level:                %u\n\n", verbosity );
    }
    
    if ( !numCallRecords )
    {
        printf ( "\nInput File is empty. "
                 ">= 1 call record is required.\n\n" );
        
        CleanupAndExit ( EXIT_FAILURE );        // control does not return here
    }
   
    printf ( "\nThe reported %% complete relates to the number of call "
             "records in the input file for which call files have been "
             "produced and dropped into Asterisk's outgoing folder. It "
             "is not the %% of SPIT calls that the Asterisk platform has "
             "successfully dialed or completed."
             "\n\nThis program is done when a call file for each call record "
             "in the input file has been dropped into Asterisk's outgoing "
             "folder.\n\n" ); 

    printf ( "\n%u%% Complete", 0 );
    fflush ( stdout );

    rewind ( hFileOfCallRecords );  //  Prepare to scan the input file again.
    
    bLookingForBlankLine = false;  //  Want to find the 1st call record in the file.

    numSPITCallsProvoked = 0;      //  No call files have been produced from the input file

    //  Main loop
    //
    //  Process all call records in the input file until a call file has been produced
    //  from each call record and dropped into Asterisk's outgoing spool folder. Before the 
    //  1st call record is processed, and after each call record produces a call file that is
    //  dropped into the outgoing folder, check to see if the outgoing folder is saturated with
    //  spitter call files.
    
    while ( numSPITCallsProvoked < numCallRecords )
    {    
        if ( spitCallsAtLimit () )
        {
            usleep ( 10000000 );   // sleep 10 seconds
        }
        else
        {
            //  Scan for the next call record, produce a correponding call file,
            //  and drop that call file into Asterisk's outgoing spool folder
            
            produceCallFileFromNextCallRec ();
                        
            printf ( "\r%d%% Complete",
                    ( ( numSPITCallsProvoked * 100 ) / numCallRecords) );
                
            fflush ( stdout );
/*            
            //  Prevent runaway spitter during spitter development 
        
            if ( ( numSPITCallsProvoked > 100 ) ||
                 ( numSPITCallsProvoked > numCallRecords ) )
            {            
                printf( "\n\nWarning: Spitter might be running away during testing. "
                        "Limiting call files generated to: %u\n",
                        ( numSPITCallsProvoked < 100 ) ? numSPITCallsProvoked : 100 );
                
                CleanupAndExit( EXIT_FAILURE );
            }
*/
        }
    } // end while ( Not Complete )
        
    printf ( "\n" );
    
    CleanupAndExit ( EXIT_SUCCESS );
    
}  //  end main


//-----------------------------------------------------------------------------
//  spitCallsAtLimit
//
//  Notifies the caller if the number of SPIT Calls in 
//  Asterisk's outgoing folder are >= the configured limit
//  spitter is permitted to contribute.
//
//  Note: The number of spitter call files should really
//             never be over the limit unless there are 
//             multiple instances of spitter running or
//             there are residual spitter files remaining in
//             Asterisk's outgoing folder in excess of a
//             prior execution of spitter. Since the limit
//             is programmable from the command line,
//             it is reasonable to conclude that the number
//             of residual spitter files might be over the
//             limit of a prior invocation of spitter, or perhaps
//             a prior invocation was executed without the
//             optional limit.
//
//  If the number SPIT generated call files >= the 
//  limit, this routine also evaluates how long the 
//  spitter has been stalled and warns the user as
//  needed.
//
//  ret:  0 = FALSE (i.e. # of SPIT calls not at the limit)
//          -1 = TRUE  (i.e. # SPIT calls at limit)
//-----------------------------------------------------------------------------

int spitCallsAtLimit ()
{
    if ( !limSpitterConcurrentCalls )
    {
        return ( 0 );  // (i.e. Calls not at limit because 0 = UNLIMITED)
    }
    
    //  Are the number of SPIT calls active or waiting within Asterisk's outgoing folder
    //  at spitter's configured limit? Call files within Asterisk's outgoing folder are counted.
    //  All files beginning with the text "spitter_call" are included.
    //
    //  Note: Since this evaluation keys on the prefix of spitter call file names, the count 
    //             naturally includes all call files produced by this instance of spitter and 
    //             all prior instances of spitter. It also counts files produced by other concurrent
    //             instances of spitter, however use of concurrent instances of spitter is
    //             discouraged. No synchronization facilities to effectively coordinate multiple
    //             instances of spitter have been implemented. Multiple concurrent instances
    //             of spitter will very likely produce a number of SPIT call files in excess of an
    //             overall intended limit.
    //
    
    struct dirent **ppsFileNameList;
        
    unsigned int numOutgoingDirFiles            = 0;
    unsigned int numOutgoingDirSpitterFiles     = 0;
    
    long spitSaturationDeltaTime                = 0;
    
    int i;

    //  Load the pointer list of filenames with the names of the files in Asterisk's
    //  outgoing directory.
    
    if ( bTestMode )
    {
        //  In test mode, an outgoing folder and a tmp folder are required
        //  to exist in the current directory. call files are produced in the
        //  pseudo tmp folder and dropped into the pseudo outgoing folder.
        
        numOutgoingDirFiles =
            scandir ( "outgoing/", &ppsFileNameList, 0, 0 );
    }
    else
    {
        numOutgoingDirFiles =
            scandir ( "/var/spool/asterisk/outgoing/", &ppsFileNameList, 0, 0 );
    }

/*    
    for ( i = 0; i < numOutgoingDirFiles; i++ )
    {
        printf ( "\n %s", ppsFileNameList[i]->d_name );
    }
*/
    
//    printf ( "\n\nPattern Match test\n" );

    //  Count number of spitter files in Asterisk's outgoing spool folder
    
    for ( i = 0; i < numOutgoingDirFiles; i++ )
    {
        if ( !fnmatch ( "spitter_call_*", ppsFileNameList[i]->d_name, 0 ) )
        {
//            printf ( "\n %s", ppsFileNameList[i]->d_name );
            numOutgoingDirSpitterFiles++;
        }
    }    
    
    free ( ppsFileNameList );
    
    if ( numOutgoingDirSpitterFiles >= limSpitterConcurrentCalls )
    {
        //  Outgoing folder is saturated with spitter files.
        
        //  Monitor how long that condition has persisted
        //  and notify the spitter admin as needed
        
        gettimeofday( &tv, NULL );
        
        if ( bPreExistingSpitterSaturation )
        {
            //  Compute delta time from last Saturation condition warning
            
            spitSaturationDeltaTime = tv.tv_sec - todSpitLastSaturationWarning;
            
            if ( spitSaturationDeltaTime > spitSaturationWarningPeriod )
            {
                printf ( "\n\nWarning: production of SPIT calls stalled for %d minutes\n",
                        ( tv.tv_sec - todSpitCallLimitReached ) / (long ) 60 );
                
                fflush( stdout );
                
                todSpitLastSaturationWarning = tv.tv_sec;
            }
        }
        else
        {
            //  The next time this routine is executed, saturation will be noted as
            //  a pre-existing condition. Record the time-of-day this saturation interval
            //  was first noted and seed the generation of a periodic warning to the
            //  the spitter admin - should the spitter be stalled for a prolonged time
            //  interval.
            
            bPreExistingSpitterSaturation = true;
            
            todSpitCallLimitReached = tv.tv_sec; //  Record TOD saturation was noted.
            
            todSpitLastSaturationWarning = tv.tv_sec;  //  Seed the periodic warning
        }
        
        return ( -1 ); // spitter calls are >= limit threshold
    }
    else
    {
        bPreExistingSpitterSaturation = false;
        
        return ( 0 ); //  spitter calls are not at the limt
    }
    
}  // end spitCallsAtLimit


//-----------------------------------------------------------------------------
//  produceCallFileFromNextCallRec
//
//  Scan for the next call record, produce a correponding
//  call file, and drop that call file into Asterisk's outgoing
//  spool folder
//
//  It is the caller's responsibility to assure there is
//  at least one more call record in the input file to be
//  found and processed!!!
//
//-----------------------------------------------------------------------------

void produceCallFileFromNextCallRec ()
{
    //  Lines of the input file are processed until the next call record has been
    //  found, a call file has been produced from it, and the file has been dropped
    //  into the Asterisk outgoing spool folder. At that point we explicitly return
    //  to the calling program

    while ( fgets ( line, 2048, hFileOfCallRecords ) != NULL )
    {       
        if ( line[0] == '\n' )   // found a blank line?
        {            
            if ( bLookingForBlankLine )
            {
                //  We scan for a blank line only when we have previously found
                //  the start of a call record, and we are copying lines from that
                //  call record into a call file. Encountering a blank line means
                //  all the lines for that call record have been copied into the 
                //  call file in the /tmp direcory. The file is ready to be closed and
                //  dropped (i.e. moved) from the /tmp directory into Asterisk's
                //  outgoing folder.
                
                closeAndDropFile ();
                                
                bLookingForBlankLine = false;  // looking for start of next call record

                return;
            }
            else
            {
                //  We only scan for a non-blank line when we are searching
                //  for the 1st line of the next call record in the input file
                
                ; //  continue scanning for start of next call record
            }
        }
        else   //  Non-blank line encountered
        {
            if ( !bLookingForBlankLine )
            {
                //  The first line of a new call record has been encountered
                
                //  Create a new random file name.
                
                tmpFileName[0] = '\0';
                
                sprintf ( tmpFileName, "spitter_call_%u.call", rand() );
                
                tmpFilePath[0] = '\0';
                
                if ( bTestMode )
                {
                    //  In test mode, an outgoing folder and a tmp folder are required
                    //  to exist in the current directory. call files are produced in the
                    //  pseudo tmp folder and dropped into the pseudo outgoing folder.

                    sprintf ( tmpFilePath, "tmp/%s", tmpFileName );
                }
                else
                {
                    sprintf ( tmpFilePath, "/tmp/%s", tmpFileName );
                }
                
                //  Create a new temporary file with the random name
                
                hTmpFile = fopen ( tmpFilePath, "w" );
                                
                //  Copy the 1st line of the present call record into that file.
                
                fputs ( line, hTmpFile );
                
                bLookingForBlankLine = true;  //  scan for end of this call record
            }
            else
            {
                //  Copy the present line into the latest /tmp file and continue scanning
                //  for the end of the present call record.
                
                fputs ( line, hTmpFile );
            }
        }
    } // end while ( scanning for next call record )
    
    if ( bLookingForBlankLine )
    {
        //  We were processing the last call record in the file and scanning
        //  for the end of that call record when EOF occurred.
        
        //  Must complete processing of last call record.
        
        closeAndDropFile ();

        return;
    }
    
    //  It is an error to reach this point. This routine should only be called if there
    //  is at least one call record within the input file remaining to be processed.
    //  If we get here, we encounted EOF on the input file while we were still 
    //  scanning for the start of a call record.
    
    printf( "\n\nError: Fell out of loop while searching for next "
            "call record in input file! spitter terminates abnormally!\n" );
    
    CleanupAndExit( EXIT_FAILURE );
    
} //  end produceCallFileFromNextCallRec


//-----------------------------------------------------------------------------
//  closeAndDropFile
//
//  Scan for the next call record, produce a correponding
//  call file, and drop that call file into Asterisk's outgoing
//  spool folder
//
//-----------------------------------------------------------------------------

void  closeAndDropFile ()
{
    fclose( hTmpFile );
                
    hTmpFile = NULL;  //  not strictly required, but good form.
                
    // Move the tmp file to Asterisk's outgoing call folder
        
    outgoingFilePath[0] = '\0';

    if ( bTestMode )
    {
        //  In test mode, an outgoing folder and a tmp folder are required
        //  to exist in the current directory. call files are produced in the
        //  pseudo tmp folder and dropped into the pseudo outgoing folder.
        
        sprintf( outgoingFilePath, "outgoing/%s", tmpFileName );
    }
    else
    {
        sprintf( outgoingFilePath,
                 "/var/spool/asterisk/outgoing/%s",
                 tmpFileName );
    }
        
    rename( tmpFilePath, outgoingFilePath ); //  (i.e. move the file )

    numSPITCallsProvoked++;                    
    
} //  end CloseAndDropFile



//-----------------------------------------------------------------------------
//  catch_signals
//
//  signal catcher and handler
//  arg1: (int) signal number
//  ret:  none, we exit the program from here
//-----------------------------------------------------------------------------

void  catch_signals (int signo)
{
    switch (signo)
    {
        case	SIGINT:
        case	SIGTERM:
        {
            printf("\nexiting...\n");
            CleanupAndExit( EXIT_SUCCESS );
        }
    }
} // end catch_signals


//-----------------------------------------------------------------------------
// CleanupAndExit
//
//  arg1: (int) status to report  i.e. EXIT_SUCCESS or EXIT_FAILURE
//  ret:  the status (or nothing in the event the call to this funtion
//           is redundant
//-----------------------------------------------------------------------------

void  CleanupAndExit ( int status )
{    

    if ( hFileOfCallRecords )
    {
        fclose ( hFileOfCallRecords );
    }
    
    if ( hTmpFile )
    {
        fclose ( hTmpFile );
    }
    
    exit( status );
} // end CleanupAndExit

//-------------------------------------------------------------------------------
// Explanation - Explain mechanics of spitter 
// 
// ret:  none
//-------------------------------------------------------------------------------

void  Explanation ()
{
    printf ( "\nThe Asterisk IP PBX must be installed on the host from"
             "\nwhich the spitter tool is executed (unless you specify the"
             "\nthe test mode option on the command line - see later)."
             "\nThe spitter tool requires read permission to the directory"
             "\ncontaining the input file (& the input file itself). It"
             "\nrequires r/w permission to the /tmp directory. It requires"
             "\nr/w permission to Asterisk's /var/spool/asterisk/outgoing/"
             "\ndirectory. In the latter case, this probably means you need"
             "\nto execute spitter as root." 
             "\n"
             "\nThe input file of SPIT targets contains one or more Asterisk"
             "\nASCII call \"records\". spitter processes the \"records\" in"
             "\norder beginning at the top of the file. spitter creates a file"
             "\nin the /tmp directory for the ith \"record\". The filename"
             "\nhas a random component. spitter merely copies the lines"
             "\ncomprising the record into the file. It closes the file and"
             "\nmoves it to Asterisk's outgoing directory. When all call"
             "\noperations related to the call file are completed, Asterisk"
             "\nremoves the call file."
             "\n"
             "\nBefore each call record is processed, if the limit option"
             "\nwas entered on the command line, Asterisk's outgoing"
             "\nspool folder is searched for the number of files matching"
             "\nthe filter: \"spitter_call*\". If the number of files found"
             "\nare >= the limit option, spitter stalls. It checks the"
             "\nnumber of files every 10 seconds. It occasionally warns the"
             "\nspitter admin of the stalled condition. Please note that"
             "\nthe design of spitter does not take into account other"
             "\nconcurrently executing instances of spitter. However it"
             "\ndoes take into account the fact that spitter call files"
             "\nproduced by a prior invocation of spitter may still exist"
             "\nin Asterisk's outgoing spool folder. When the number of"
             "\nspitter call files in the outgoing folder drops below"
             "\nthe optional limit, spitter resumes dropping call files"
             "\nuntil call records in the input file are exhausted or"
             "\nthe outgoing folder is saturated with spitter call files"
             "\nonce again. Use of the limit option is highly recommended."
             "\n"
             "\nTo spitter, each call record is simply a series of non-blank"
             "\nlines. Records are separated by a line beginning with"
             "\nthe new line character (i.e. \\n = 0x0a). spitter simply"
             "\ncreates a separate file for each record. spitter does not"
             "\nparse the lines in a record to confirm it is complete or"
             "\nmakes sense."
             "\n"
             "\nIF YOUR INPUT FILE HAS MANY CALL RECORDS, BUT YOU ARE"
             "\nONLY SEEING 1 CALL LAUNCHED, IT'S LIKELY THAT YOUR CALL"
             "\nRECORDS ARE SEPARATED BY CRLF (i.e. 0x0d0a) AND NOT"
             "\nSIMPLY LF (i.e. 0x0a)."
             "\n"
             "\nYou may use the optional test mode to play with spitter"
             "\nin the absence of an Asterisk installation or simply to"
             "\ntest changes you might make to the program. In test mode"
             "\nyou must have created a tmp directory and an outgoing" 
             "\ndirectory within your current directory. spitter creates"
             "\ncall files in your personal /tmp directory and moves them"
             "\ncontingent upon the optional limit into your personal"
             "\n/outgoing directory. Naturally, if a limit is specified"
             "\non the command line, you may need to remove call files"
             "\npreviously dropped into your personal outgoing directory"
             "\nor spitter may stall - unless it is your intent to test"
             "\nspitter's stall logic."
             "\n\n" );

}  //  end Explanation


//-------------------------------------------------------------------------------
// usage 
// 
// ret:  none
//-------------------------------------------------------------------------------

void  usage ()
{
    printf ( "\n%s", __SPITTER_VERSION );
    printf ( "\n%s", __SPITTER_DATE );
    
    printf ( "\n Usage:"                                                                     );
    printf ( "\n Mandatory -"                                                                );
    printf ( "\n\tfile of spit call records"                                                 );
    printf ( "\n Optional -"                                                                 );
    printf ( "\n \t-l limit on max # spitter's calls in Asterisk's outgoing folder."         );
    printf ( "\n \t   spitter can be throttled by this optional parm."                       );
    printf ( "\n \t   (e.g. -l 10 means 10 concurrent spit calls) [default = 0 = unlimited]" );
    printf ( "\n \t   Note: this is the limit on spitter's contribution,"                    );
    printf ( "\n \t         not a limit on the total # of outgoing calls!"                   );
    printf ( "\n \t-t test mode (e.g. skips check that Asterisk is running and"              );
    printf ( "\n \t         presumes you have a tmp folder and an outgoing folder"           );
    printf ( "\n \t         in your current directory)"                                      );
    printf ( "\n \t-v verbose output mode (from -v up to -vvv)"                              );
    printf ( "\n \t-h help & explanation\n"                                                  );
    printf ( "\n\n"                                                                          );

    exit ( EXIT_FAILURE );     // control does not return here
} // end usage
